/*
 * (C) Copyright 2008-2010 Jean-Christophe PLAGNIOL-VILLARD <plagnioj@jcrosoft.com>
 *
 * See file CREDITS for list of people who contributed to this
 * project.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; version 2 of
 * the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 */

#include <common.h>
#include <digest.h>
#include <malloc.h>
#include <fs.h>
#include <fcntl.h>
#include <linux/stat.h>
#include <errno.h>
#include <module.h>
#include <linux/err.h>

static LIST_HEAD(digests);

static struct digest_algo* digest_algo_get_by_name(const char *name);

static int dummy_init(struct digest *d)
{
	return 0;
}

static void dummy_free(struct digest *d) {}

int digest_algo_register(struct digest_algo *d)
{
	if (!d || !d->name || !d->update || !d->final || d->length < 1)
		return -EINVAL;

	if (!d->init)
		d->init = dummy_init;

	if (!d->alloc)
		d->alloc = dummy_init;

	if (!d->free)
		d->free = dummy_free;

	if (digest_algo_get_by_name(d->name))
		return -EEXIST;

	list_add_tail(&d->list, &digests);

	return 0;
}
EXPORT_SYMBOL(digest_algo_register);

void digest_algo_unregister(struct digest_algo *d)
{
	if (!d)
		return;

	list_del(&d->list);
}
EXPORT_SYMBOL(digest_algo_unregister);

static struct digest_algo *digest_algo_get_by_name(const char *name)
{
	struct digest_algo* d;

	if (!name)
		return NULL;

	list_for_each_entry(d, &digests, list) {
		if(strcmp(d->name, name) == 0)
			return d;
	}

	return NULL;
}

struct digest *digest_alloc(const char *name)
{
	struct digest* d;
	struct digest_algo* algo;

	algo = digest_algo_get_by_name(name);
	if (!algo)
		return NULL;

	d = xzalloc(sizeof(*d));
	d->algo = algo;
	d->ctx = xzalloc(algo->ctx_length);
	if (d->algo->alloc(d)) {
		digest_free(d);
		return NULL;
	}

	return d;
}
EXPORT_SYMBOL_GPL(digest_alloc);

void digest_free(struct digest *d)
{
	if (!d)
		return;
	d->algo->free(d);
	free(d->ctx);
	free(d);
}
EXPORT_SYMBOL_GPL(digest_free);

int digest_file_window(struct digest *d, const char *filename,
		       const unsigned char *key, size_t keylen,
		       unsigned char *hash,
		       ulong start, ulong size)
{
	ulong len = 0;
	int fd, now, ret = 0;
	unsigned char *buf;
	int flags = 0;

	if (key)
		digest_set_key(d, key, keylen);

	digest_init(d);

	fd = open(filename, O_RDONLY);
	if (fd < 0) {
		perror(filename);
		return fd;
	}

	buf = memmap(fd, PROT_READ);
	if (buf == (void *)-1) {
		buf = xmalloc(4096);
		flags = 1;
	}

	if (start > 0) {
		if (flags) {
			ret = lseek(fd, start, SEEK_SET);
			if (ret == -1) {
				perror("lseek");
				goto out;
			}
		} else {
			buf += start;
		}
	}

	while (size) {
		now = min((ulong)4096, size);
		if (flags) {
			now = read(fd, buf, now);
			if (now < 0) {
				ret = now;
				perror("read");
				goto out_free;
			}
			if (!now)
				break;
		}

		if (ctrlc()) {
			ret = -EINTR;
			goto out_free;
		}

		digest_update(d, buf, now);
		size -= now;
		len += now;
	}

	digest_final(d, hash);

out_free:
	if (flags)
		free(buf);
out:
	close(fd);

	return ret;
}
EXPORT_SYMBOL_GPL(digest_file_window);

int digest_file(struct digest *d, const char *filename,
		       const unsigned char *key, size_t keylen,
		       unsigned char *hash)
{
	struct stat st;
	int ret;

	ret = stat(filename, &st);

	if (ret < 0)
		return ret;

	return digest_file_window(d, filename, key, keylen, hash, 0, st.st_size);
}
EXPORT_SYMBOL_GPL(digest_file);

int digest_file_by_name(const char *algo, const char *filename,
		       const unsigned char *key, size_t keylen,
		       unsigned char *hash)
{
	struct digest *d;
	int ret;

	d = digest_alloc(algo);
	if (!d)
		return -EIO;

	ret = digest_file(d, filename, key, keylen, hash);
	digest_free(d);
	return ret;
}
EXPORT_SYMBOL_GPL(digest_file_by_name);
